/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg;

import static org.fortiss.af3.exploration.alg.ExplorationAlgActivator.getDefault;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import org.eclipse.ui.IStartup;
import org.fortiss.af3.exploration.alg.backend.Opt4JDseBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.MessageEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.StrictTTDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm.MessageDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoderAcyclic;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.partitionmapping.PartitionMappingDecoderGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFGraphDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.taskmapping.AbstractTaskMappingIdentityDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.taskmapping.instantiate.FailSilentAbstractTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.taskmapping.instantiate.TaskInstanceResourceAlignmentDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.service.IExplorationConstraintTransformationService;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationModuleService;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.IExplorationService;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationConstraintTransformationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationEvaluationService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationModuleService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.internal.ExplorationTransformationService;
import org.fortiss.af3.exploration.backend.DseBackendHandler;
import org.osgi.framework.BundleContext;

/**
 * Centralized class that provides the startup routines of all services provided by this plugin.
 * 
 * The order of execution is as follows:
 * - Initialize own services.
 * - Register with the exploration plugin. (Make the "Parent" aware of me)
 * - Start own services.
 * 
 * @author diewald
 */
public final class ExplorationAlg implements IStartup {

	/** {@inheritDoc} */
	@Override
	public void earlyStartup() {
		initialize();
		registerBackend();
		start();
	}

	/**
	 * Initializes the {@link IExplorationService}s contained in this plugin (first service
	 * activation stage).
	 */
	public static void initialize() {
		registerServices();

		getService(IExplorationModuleService.class).initializeService();
		getService(IExplorationConstraintTransformationService.class).initializeService();
		getService(IExplorationEvaluatationService.class).initializeService();
		getService(IExplorationRepairService.class).initializeService();
		getService(IExplorationTransformationService.class).initializeService();

		// Initializes/Registers the bundled exploration modules.
		registerExplorationModules();
	}

	/** Registers the provided services with the {@link BundleContext}. */
	private static void registerServices() {
		BundleContext context = getDefault().getBundle().getBundleContext();

		context.registerService(IExplorationModuleService.class.getName(),
				ExplorationModuleService.getInstance(), null);
		context.registerService(IExplorationConstraintTransformationService.class.getName(),
				ExplorationConstraintTransformationService.getInstance(), null);
		context.registerService(IExplorationEvaluatationService.class.getName(),
				ExplorationEvaluationService.getInstance(), null);
		context.registerService(IExplorationRepairService.class.getName(),
				ExplorationRepairService.getInstance(), null);
		context.registerService(IExplorationTransformationService.class.getName(),
				ExplorationTransformationService.getInstance(), null);
	}

	/** Registers the exploration modules bundled with this plugin. */
	@SuppressWarnings("unchecked")
	private static void registerExplorationModules() {
		// Functional groups of Decoders.
		registerSafetyFunctionModules();
		registerTaskMappingModules();
		registerInstTaskGraphModules();

		// Single exploration modules: Decoding.
		getService(IExplorationModuleService.class).registerProblem(StrictTTSchedule.class,
				new StrictTTDecoder());
		getService(IExplorationModuleService.class).registerProblem(
				PlatformCommunicationGraphEncoding.class, new PartitionMappingDecoderGraph());
		getService(IExplorationModuleService.class).registerProblem(MessageEncoding.class,
				new MessageDecoder());
	}

	/** Registers problem modules related to safety functions. */
	@SuppressWarnings("unchecked")
	private static void registerSafetyFunctionModules() {
		getService(IExplorationModuleService.class).registerProblem(
				AbstractTaskMappingEncoding.class, new SFMappingDecoder());
		// TODO: ConstraintDecoder
		// getService(IExplorationModuleService.class).registerProblem(
		// AbstractTaskMappingEncoding.class, new SFMappingConstraintDecoder());
		getService(IExplorationModuleService.class).registerProblem(SafeTaskGraphEncoding.class,
				new SFGraphDecoder());
	}

	/** Registers all problem modules related to task mapping. */
	private static void registerTaskMappingModules() {
		getService(IExplorationModuleService.class).registerProblem(
				AbstractTaskMappingEncoding.class, new AbstractTaskMappingIdentityDecoder());
		getService(IExplorationModuleService.class).registerProblem(
				FailSilentTaskMappingEncoding.class, new FailSilentAbstractTaskMappingDecoder());
		// getService(IExplorationModuleService.class).registerProblem(
		// FaultDetectionVotingTaskMappingEncoding.class,
		// new FaultDetectionVotingAbstractTaskMappingDecoder());
		getService(IExplorationModuleService.class).registerProblem(
				InstantiatedTaskMappingEncoding.class, new TaskInstanceResourceAlignmentDecoder());
	}

	/**
	 * Registers all modules related to the derivation of task graphs from instantiated task
	 * mappings.
	 */
	@SuppressWarnings("unchecked")
	private static void registerInstTaskGraphModules() {
		getService(IExplorationModuleService.class).registerProblem(
				InstantiatedTaskGraphEncoding.class, new InstantiatedTaskMappingDecoder());
		getService(IExplorationModuleService.class).registerProblem(
				InstantiatedAcyclicTaskGraphEncoding.class,
				new InstantiatedTaskMappingDecoderAcyclic());
	}

	/** Register the DSE backend with the exploration plugin. */
	private static void registerBackend() {
		DseBackendHandler.getInstance().registerDseBackend(new Opt4JDseBackend());
	}

	/**
	 * Starts the {@link IExplorationService}s contained in this plugin (second service activation
	 * stage).
	 */
	public static void start() {
		// TODO: Check start order w.r.t. the transformation service.
		getService(IExplorationModuleService.class).startService();
		getService(IExplorationConstraintTransformationService.class).startService();
		// The Transformation Service must be started before the Evaluation Service such that it can
		// check for the availability of required transformation modules.
		getService(IExplorationTransformationService.class).startService();
		getService(IExplorationEvaluatationService.class).startService();
		getService(IExplorationRepairService.class).startService();
	}
}
