/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping;

import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.opt4j.core.Genotype;

/**
 * {@link Genotype} which encodes an {@link TaskMappingEncoding} that may contain
 * {@link TaskMappingEntry}s which are required to be initialized.
 * 
 * @author diewald
 */
public class AbstractTaskMappingEncoding extends
		TaskMappingEncoding<ITaskAdapter<?>, AbstractTaskMappingEntry> {

	/**
	 * Constructor that is initialized with a reference to the DSE specification and the input
	 * system model.
	 */
	AbstractTaskMappingEncoding(PartitionMappingEncoding partEnc, SafeTaskGraphEncoding tgEncoding) {
		super(partEnc, tgEncoding);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	public AbstractTaskMappingEncoding(PartitionMappingEncoding partEnc,
			SafeTaskGraphEncoding tgEncoding, boolean initialize) throws ExplorationException {
		super(partEnc, tgEncoding, initialize);
	}

	/** Copy Constructor. */
	public AbstractTaskMappingEncoding(PartitionMappingEncoding partEnc,
			TaskGraphEncoding<?> tgEncoding) {
		super(partEnc, tgEncoding);
	}

	/**
	 * Constructor for UPDATES.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	public AbstractTaskMappingEncoding(PartitionMappingEncoding partEnc,
			AbstractTaskGraphEncoding atgEnc, boolean initialize) throws ExplorationException {
		super(partEnc, atgEnc, initialize);
	}

	/**
	 * Constructor that generated a new {@link AbstractTaskMappingEncoding} preserving the
	 * allocations from the given {@code partialEncoding}.
	 * 
	 * Constructor for UPDATES.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one {@link ITaskAdapter}.
	 */
	public AbstractTaskMappingEncoding(PartitionMappingEncoding partEnc,
			AbstractTaskGraphEncoding atgEnc, AbstractTaskMappingEncoding partialEncoding)
			throws ExplorationException {
		super(partEnc, atgEnc, partialEncoding);
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEntry createEntry(ITaskAdapter<?> deployableComponent,
			Partition deploymentTarget) {
		// Request not needed to construct a {@link FailSilentTaskMappingEncoding}
		if(deploymentTarget == null) {
			return null;
		}

		return new AbstractTaskMappingEntry(deployableComponent, deploymentTarget);
	}

	/** {@inheritDoc} */
	@Override
	protected <S extends TaskMappingEntry<ITaskAdapter<?>, Partition>> AbstractTaskMappingEntry
			createEntry(S allocEntry) {
		// TODO: Remove the ugly casts.
		return new AbstractTaskMappingEntry(allocEntry.getSource(), allocEntry.getTarget());
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public AbstractTaskMappingEncoding newInstance() {
		try {
			return new AbstractTaskMappingEncoding(null, null, true);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding clone() {
		AbstractTaskMappingEncoding clonedEncoding;
		clonedEncoding = new AbstractTaskMappingEncoding(partMappingEncoding, tgEncoding);

		Map<AbstractTaskMappingEntry, AbstractTaskMappingEntry> copyAssocMap = new HashMap<>();

		clonedEncoding.allocationMap = copyAllocationMap(copyAssocMap);
		clonedEncoding.constraintMap = copyConstraintMap(copyAssocMap);
		clonedEncoding.getRequesters().forEach(c -> c.updateInstatiationData());
		return clonedEncoding;
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposableGenotype> getGTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends Phenotype> getPTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}
}
