/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FailSilentTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingRuntimeException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationModuleService;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.start.Opt4JTask;

import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Provides;

/**
 * {@link Decoder} that handles dependencies between {@link Decoder}s, {@link Genotype}s, and
 * {@link Phenotype}s using the Guice framework.
 * 
 * @author diewald
 */
public class GuiceDecoder extends AbstractModule implements Decoder<Genotype, Phenotype> {

	/** SystemModelAdapter containing input parameters. */
	// FIXME: Move to general injector in the ExplorationModule service.
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> sysModel;

	/** Resulting phenotype from the decoding phase that is composed of sub-{@link Phenotype}s. */
	protected Phenotype composedPhenotype;

	/** Defines the desired exploration result type. */
	// TODO: Remove this hard coded line and get the desired result from the backend interface.
	protected Class<? extends Phenotype> resultType = FailSilentTaskMappingEncoding.class;

	/** Injector created by the {@link Opt4JTask}. */
	@Inject
	protected Injector opt4JInjector;

	/** Child injector for the dependency management of the decode phase. */
	protected Injector decodingInjector;

	/** Reference to the input {@link Genotype}. */
	// @ThreadScoped
	protected ThreadLocal<CompositeGenotype<Class<? extends Genotype>, ? extends Genotype>> genotype =
			new ThreadLocal<>();

	/** Constructor. */
	public GuiceDecoder(SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> sysModel) {
		this.sysModel = sysModel;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Phenotype decode(Genotype genotype) {
		// The decode operations are responsible to add the results of the decode operation to the
		// parent phenotype.
		if(!(genotype instanceof CompositeGenotype)) {
			throw new DecodingRuntimeException(genotype, this,
					"The top-most DSE process decoder requires a CompositeGenotype as input!");
		}
		this.genotype
				.set((CompositeGenotype<Class<? extends Genotype>, ? extends Genotype>)genotype);
		// TODO(AD): See configure(): We need a service machenism and user configuration to select
		// the desired set of activated decoders.
		decodingInjector = opt4JInjector.createChildInjector(this);
		decodingInjector.getInstance(resultType);
		return composedPhenotype;
	}

	/** Provides the composite {@link Genotype} as required by the Guice framework. */
	@Provides
	// @ThreadScoped
			public
			CompositeGenotype<Class<? extends Genotype>, ? extends Genotype> getGenotype() {
		return genotype.get();
	}

	/** Returns the sub-genotype of the given type, or {@code null} if none is found. */
	public <T extends Genotype> T getGenotype(Class<T> genotypeType) {
		return getGenotype().get(genotypeType);
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		// bindScope(ThreadScoped.class, new ThreadScope());
		install(sysModel);

		for(DecoderModule<?> module : getService(IExplorationModuleService.class).getDecoders()) {
			install(module);
		}
	}

	/*
	 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	 * HACKs below! They are used to retrieve the concrete genotypes. Scoping shall resolve the
	 * problem of Genotypes living in parallel.
	 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	 */
	/**
	 * Returns the {@link SafetyFunctionArchEncoding} sub-{@link Genotype} passed to this decoder.
	 * (Testing!!!)
	 */
	// TODO: Move to the encoding to render the approach more modular.
	@Provides
	public SafetyFunctionArchEncoding getSFAEnc() {
		return getGenotype(SafetyFunctionArchEncoding.class);
	}

	// @Provides
	// public SafetyFunctionArchEncoding getSFAEnc() {
	// return getGenotype(SafetyFunctionArchEncoding.class);
	// }
	//
	// @Provides
	// public SafetyFunctionArchEncoding getSFAEnc() {
	// return getGenotype(SafetyFunctionArchEncoding.class);
	// }
}
