/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.partitionmapping;

import java.util.ArrayList;
import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.PartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultEdge;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;

import com.google.inject.Provides;

/**
 * Update the {@link PlatformCommunicationGraphEncoding} with potential changes from the partition
 * mapping. Partitions are added as leaves to the graph (behind phys. execution units).
 * 
 * @author diewald
 */
// TODO: DecodeUpdate: We may consider moving the decode method to the creator.
public class PartitionMappingDecoderGraph extends DecoderModule<PlatformCommunicationGraphEncoding> {

	@Provides
	PartitionMappingEncoding getPmEnc(
			CompositeGenotype<Class<? extends Genotype>, ? extends Genotype> compGene) {
		return compGene.get(PartitionMappingEncoding.class);
	}

	@Provides
	public PlatformCommunicationGraphEncoding decodeComposable(
			PlatformCommunicationGraphEncoding pcmEnc, PartitionMappingEncoding pmEnc) {
		for(IResourceAdapter<?> execUnit : pcmEnc.getActualGraph().vertexSet()) {
			for(IResourceAdapter<?> partition : pmEnc.getRequesters(execUnit)) {
				pcmEnc.getActualGraph().addVertex(partition);
				pcmEnc.getActualGraph().addEdge(partition, execUnit);
				pcmEnc.getActualGraph().addEdge(execUnit, partition);
			}
		}

		return pcmEnc;
	}

	public PlatformCommunicationGraphEncoding updateComposable(
			PlatformCommunicationGraphEncoding pcmEnc, PartitionMappingEncoding pmEnc) {
		// Remove non-existing Partitions (from a previous iteration).
		Collection<IResourceAdapter<?>> removeVertices = new ArrayList<>();
		for(IResourceAdapter<?> resource : pcmEnc.getActualGraph().vertexSet()) {
			if((resource instanceof Partition) && !pmEnc.getRequesters().contains(resource)) {
				Collection<DefaultEdge> edgeSet = pcmEnc.getActualGraph().edgesOf(resource);
				pcmEnc.getActualGraph().removeAllEdges(edgeSet);
				removeVertices.add(resource);
			}
		}
		pcmEnc.getActualGraph().removeAllVertices(removeVertices);
		// Add missing Partitions.
		for(Partition partition : pmEnc.getRequesters()) {
			if(!pcmEnc.getActualGraph().containsVertex(partition)) {
				// We have only one Partition per entry --> Multiples for task mappings.
				// TODO: Correct the return type for single and multi returns.
				IResourceAdapter<?> execUnit =
						pmEnc.getAllocatedResourcesFor(partition).iterator().next();
				pcmEnc.getActualGraph().addVertex(partition);
				// FIXME: HACK!!! Connects the Partitions to the communication resource attached to
				// the phys exec unit. This is used to avoid larger reworks in the TDMA Message
				// scheduler. The scheduler must be reconsidered to reflect internal comm. of a
				// core/processor correctly.
				Collection<IResourceAdapter<?>> commResources =
						Graphs.neighborListOf(pcmEnc.getActualGraph(), execUnit);
				IResourceAdapter<?> commresource =
						LambdaUtils.getFirst(commResources,
								r -> r instanceof ICommunicationResourceAdapter).get();
				pcmEnc.getActualGraph().addEdge(partition, commresource);
				pcmEnc.getActualGraph().addEdge(commresource, partition);
			}
		}
		return pcmEnc;
	}

	/** {@inheritDoc} */
	@Override
	public PlatformCommunicationGraphEncoding decode(Genotype genotype) {
		// TODO Auto-generated method stub
		return null;
	}
}
